//-----------------------------------------------------------------------------
// Copyright  2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// program	linesort
//
// purpose	Perform a simple sort of input lines via string mapping.
//
// syntax	linesort < infile
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <map.h>

//-----------------------------------------------------------------------------
// Define flag bits for each line ending character that is handled.  A group
// of unlike line ending characters together is treated as a single ending.
// When any line ending character is encounted again, this indicates a new
// and empty line is present.
//-----------------------------------------------------------------------------
static unsigned char	end_table	[256] = {
    1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  2,  4,  8, 16,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
};

//-----------------------------------------------------------------------------
// function	fgetline_copy
//
// purpose	Get input from an open FILE one line at a time, copied to a
//		destination buffer, with length limiting.
//
// arguments	1 (FILE *) file stream pointer
//		2 (char *) destination buffer
//		3 (size_t) maximum length of buffer
//
// note		One fewer character than the specified maximum length will
//		be copied to allow room for the termination character.
//
// returns	(size_t) actual length of line (even if not all copied)
//		(size_t)  0 = empty line (normal condition)
//		(size_t) ~0 = EOF, no more lines, or error
//
// note		The length returned is the full length of the line even if
//		fewer characters are copied to the given buffer.  The calling
//		program will need to perform its own checks for truncated
//		line data where that is needed.
//-----------------------------------------------------------------------------
size_t
fgetline_copy (
    FILE *	arg_file
    ,
    char *	arg_buffer
    ,
    size_t	arg_length
)
{
    size_t	len_count	;
    int		end_flags	;
    int		ch_code		;

    //-- Check for a hoser calling us.
    if ( ! arg_file ) return ~0U;
    if ( ! arg_buffer ) return ~0U;

    //-- Check for no available input.
    if ( feof( arg_file ) ) return ~0U;
    if ( ferror( arg_file ) ) return ~0U;

    //-- Initialize.
    len_count = 0;
    end_flags = 0;

    for (;;) {

	//-- Get next character and check for EOF.
	if ( ( ch_code = fgetc( arg_file ) ) == EOF ) break;

	//-- If this character is a line end ...
	if ( end_table[ch_code] ) {

	    //-- If already seen then this is another line.
	    if ( end_flags & end_table[ch_code] ) break;

	    //-- Remember this one has been seen.
	    end_flags |= end_table[ch_code];

	} else {	//-- This is a data character.

	    //-- If line end seen then this is another line.
	    if ( end_flags ) break;

	    //-- Count it even if we do not store it.
	    ++ len_count;

	    //-- If buffer space remains, store the character.
	    if ( arg_length > 1 ) {
		* arg_buffer = ch_code;
		++ arg_buffer;
		-- arg_length;
	    }
	}
    }

    //-- Put back the start of the next line.
    if ( ch_code != EOF ) ungetc( ch_code, arg_file );

    //-- Terminate the stored line.
    * arg_buffer = 0;

    //-- All done: return real length to caller.
    return len_count;
}

//-----------------------------------------------------------------------------
// function	main
//-----------------------------------------------------------------------------
int
main (
    int		argc
,
    char * *	argv
)
{
    char		line_buf	[4096]	;
    const char *	line_ptr	;
    MAP			line_map	;
    size_t		line_len	;


    //--------------------
    // Create the mapping.
    //--------------------
    line_map = map_uca_new( NULL );

    //-----------------------------------------------
    // Read each line and insert it into the mapping.
    //-----------------------------------------------
    while ( ( line_len = fgetline_copy( stdin, line_buf, 4096 ) ) != ~0 ) {
	if ( line_len > 4095 ) line_len = 4095;
	if ( map_uca_find_exact( line_map, line_buf, line_len ) < 0 ) {
	    map_uca_insert( line_map, line_buf, line_len );
	}
    }
    
    //----------------------------------------------
    // Write every line from the mapping.
    //----------------------------------------------
    map_unselect( line_map );
    while ( map_forward( line_map ) >= 0 ) {
	line_ptr = map_uca_key_ptr( line_map, NULL );
	puts( line_ptr );
    }

    //------
    // Poof.
    //------
    map_destroy( line_map );

    return 0;
}
