//-----------------------------------------------------------------------------
// Copyright  2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		map_scalar_lib.h
//
// purpose	Define common functions for scalar data type implementation.
//-----------------------------------------------------------------------------

#include "map_lib.h"

//-----------------------------------------------------------------------------
// The following must be defined by the file that includes this one:
//
// map_type		Typedef this to the C type for the scalar type to be
//			used for the key or data, e.g. unsigned long.
//
// MAP_TYPE_ID		Define this to the symbolic name for this type, as
//			used in the data mix union, e.g. si.
//
// MAP_TYPE_CODE	Define this to the int code value that identifies
//			this data type in a mixed data struct.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// struct	map_link
// struct	map_node
//
// purpose	Define structures needed to keep the AVL trees used to hold
//		the mapping.
//-----------------------------------------------------------------------------
struct map_link {
    avl_link		link	;
    map_type		key	;
};
struct map_node {
    struct map_mixed	data	;
    struct map_link	link	;
};

//-----------------------------------------------------------------------------
// function	map_xx_compare (internal)
//
// purpose	Compare the scalar keys in two AVL tree nodes.  This function
//		is not globally exported.  It is used only to establish the
//		comparison method for the AVL tree used to hold this mapping.
//
// arguments	1 (struct map_link *)
//		2 (struct map_link *)
//		3 (void *) extra pointer (not used)
//
// returns	(int) comparison status, -1, 0, 1
//-----------------------------------------------------------------------------
static
int
map_xx_compare (
    avl_link *		arg_one
    ,
    avl_link *		arg_two
    ,
    AVL *		arg_avl
) {

#ifdef MAP_COMPARE_CAST
    return ( ((MAP_COMPARE_CAST) (((struct map_link *)arg_one)->key))
	     < ((MAP_COMPARE_CAST) (((struct map_link *)arg_two)->key)) ) ?
	-1 : ( ((MAP_COMPARE_CAST) (((struct map_link *)arg_one)->key))
	       > ((MAP_COMPARE_CAST) (((struct map_link *)arg_two)->key)) ) ? 1 : 0;
#else
    return ( ((struct map_link *)arg_one)->key < ((struct map_link *)arg_two)->key ) ?
	-1 : ( ((struct map_link *)arg_one)->key > ((struct map_link *)arg_two)->key ) ? 1 : 0;
#endif
}

//-----------------------------------------------------------------------------
// function	map_xx_compare_alt (internal)
//
// purpose	Compare the scalar keys in two AVL tree nodes.  This function
//		is not globally exported.  It is used only to establish the
//		comparison method for the AVL tree used to hold this mapping.
//
// arguments	1 (struct map_link *)
//		2 (struct map_link *)
//		3 (void *) alternate compare function
//
// returns	(int) comparison status, -1, 0, 1
//-----------------------------------------------------------------------------
static
int
map_xx_compare_alt (
    avl_link *		arg_one
    ,
    avl_link *		arg_two
    ,
    AVL *		arg_avl
) {
    if ( map_cmp_func( arg_avl ) ) {
	return ( * map_cmp_func( arg_avl ) )
	    ( ((struct map_link *)arg_one)->key, ((struct map_link *)arg_two)->key );
    }
    return ( ((struct map_link *)arg_one)->key < ((struct map_link *)arg_two)->key ) ?
	-1 : ( ((struct map_link *)arg_one)->key > ((struct map_link *)arg_two)->key ) ? 1 : 0;
}

//-----------------------------------------------------------------------------
// function	map_xx_new
//
// purpose	Create a new mapping with the specific scalar key type.
//
// arguments	1 (int *)() optional alternate comparison function
//
// returns	(MAP) handle for the new mapping.
//		(MAP) NULL = error creating new mapping
//-----------------------------------------------------------------------------
MAP
map_xx_new (
    int		(* arg_compare)(map_type,map_type)
)
{
    MAP		map_p		;

    //-- Allocate a structure for the root of the mapping.
    if ( ! ( map_p = malloc( sizeof (struct map_root) ) ) ) return NULL;

    //-- Record the key type.
    map_p->key_type = MAP_TYPE_CODE;

    //-- Let the AVL layer initialize the rest of it.
    if ( arg_compare ) {
	map_cmp_func( map_p ) = arg_compare;
	if ( ! avl_init( & (map_p->avl_tree), struct map_node, link, map_xx_compare_alt ) ) return NULL;
    } else {
	if ( ! avl_init( & (map_p->avl_tree), struct map_node, link, map_xx_compare ) ) return NULL;
    }

    //-- Return the handle to the new mapping.
    return map_p;
}

//-----------------------------------------------------------------------------
// function	map_xx_find
//
// purpose	Find a node with a matching key, or possibly a nearby key.
//
//		If the influential direction is none, return the first exact
//		match encountered, or return none.
//
//		If the influential direction is low, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest node in the
//		low direction.
//
//		If the influential direction is high, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest node in the
//		high direction.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key
//		3 (int) direction and near option, -2, -1, 0, 1, 2
//			-2 = choose lowest exact, or nearest lower
//			-1 = choose lowest exact only
//			 0 = choose first exact only
//			 1 = choose highest exact only
//			 2 = choose highest exact, or nearest higher
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find (
    MAP			arg_map
    ,
    map_type		arg_key
    ,
    int			arg_dir_near
)
{
    struct map_node *	node_p	;
    struct map_link	finder	;

    //-- Make sure we have a correct mapping.
    if ( ! arg_map ) return -2;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return -2;

    //-- Set up a search link to pass the key.
    finder.key = arg_key;

    //-- Let the AVL layer do the search.
    node_p = avl_find( & (arg_map->avl_tree), & finder, arg_dir_near );

    //-- Return the found data type or -1 if none found.
    return ( node_p ) ? node_p->data.type : -1;
}

//-----------------------------------------------------------------------------
// function	map_xx_find_exact
//
// purpose	Find a member with an exactly matching key of the specific
//		array key type for this mapping.  If multiple members match,
//		this function does not define what particular member will be
//		selected.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_exact (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    return map_xx_find( arg_map, arg_key, 0 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_exact_lo
//
// purpose	Find a node with an exactly matching key of the specific
//		scalar key type for this mapping.  If there is more than
//		one node of the same key, select the one in the lowest
//		position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_exact_lo (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    return map_xx_find( arg_map, arg_key, -1 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_lo_or_near
//
// purpose	Find a node with an exactly matching key, or if no exactly
//		matching key present, a node with nearest lower value, of
//		the specific scalar key type for this mapping.  If there is
//		more than one node of the same key, select the one in the
//		lowest position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_lo_or_near (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    return map_xx_find( arg_map, arg_key, -2 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_exact_hi
//
// purpose	Find a node with an exactly matching key of the specific
//		scalar key type for this mapping.  If there is more than
//		one node of the same key, select the one in the highest
//		position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_exact_hi (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    return map_xx_find( arg_map, arg_key, 1 );
}

//-----------------------------------------------------------------------------
// function	map_xx_find_hi_or_near
//
// purpose	Find a node with an exactly matching key, or if no exactly
//		matching key present, a node with nearest higher value, of
//		the specific scalar key type for this mapping.  If there is
//		more than one node of the same key, select the one in the
//		highest position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_hi_or_near (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    return map_xx_find( arg_map, arg_key, 2 );
}

//-----------------------------------------------------------------------------
// function	map_xx_insert
//
// purpose	Insert a new member into a mapping with the specified key and
//		an initial undefined data value, unless one already exists
//		with that exact key, in which case it is made active.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int)  0 = found or inserted a member with no data
//		(int) >0 = data type of found node
//-----------------------------------------------------------------------------
int
map_xx_insert (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    struct map_node *	node_p	;
    int			rc	;

    //-- Make sure we have a correct mapping.
    if ( ! arg_map ) return -2;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return -2;

    //-- Allocate a combined node and key space.  The key is a scalar
    //-- so there is always just one element and it is in the node.
    if ( ! ( node_p = malloc( sizeof (struct map_node) ) ) ) return -2;

    //-- Initialize the node.
    node_p->data.type = 0;
    node_p->link.key = arg_key;

    //-- Insert node or return data type if already exists.
    rc = avl_insert( & (arg_map->avl_tree), node_p );
    if ( rc >= 0 ) return 0;
    free( node_p );
    return map_xx_find( arg_map, arg_key, 0 );
}

//-----------------------------------------------------------------------------
// function	map_xx_insert_dup
//
// purpose	Insert a new node into a mapping with the specified key and
//		an initial undefined data value.  If a node already exists,
//		insert a duplicate.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = not inserted
//		(int)  1 = new unique node inserted
//		(int)  2 = duplicate node inserted
//-----------------------------------------------------------------------------
int
map_xx_insert_dup (
    MAP			arg_map
    ,
    map_type		arg_key
)
{
    struct map_node *	node_p	;
    int			rc	;

    //-- Make sure we have a correct mapping.
    if ( ! arg_map ) return -2;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return -2;

    //-- Allocate a combined node and key space.  The key is a scalar
    //-- so there is always just one element and it is in the node.
    if ( ! ( node_p = malloc( sizeof (struct map_node) ) ) ) return -2;

    //-- Initialize the node.
    node_p->data.type = 0;
    node_p->link.key = arg_key;

    //-- Insert the node.
    rc = avl_insert_dup( & (arg_map->avl_tree), node_p );

    //-- Return status based on insertion.
    return ( rc == -1 ) ? 0 : rc;
}

//-----------------------------------------------------------------------------
// function	map_xx_key
//
// purpose	Get the key value of the current node in the mapping.
//
// arguments	1 (MAP) mapping handle
//
// return	(map_type) key value
//		(map_type) 0 cast to the key type of the tree
//-----------------------------------------------------------------------------
map_type
map_xx_key (
    MAP			arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a correct mapping and current node.
    if ( ! arg_map ) return (map_type) 0;
    if ( arg_map->key_type != MAP_TYPE_CODE ) return (map_type) 0;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return (map_type) 0;

    //-- Return the key.
    return node_p->link.key;
}

//-----------------------------------------------------------------------------
// function	map_xx_store
//
// purpose	Store the given data into the current node in the mapping.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) data value to store
//
// return	(int) -2 = error, no map, no node
//		(int)  0 = success, data stored
//-----------------------------------------------------------------------------
int
map_xx_store (
    MAP			arg_map
    ,
    map_type		arg_data
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return -2;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return -2;

    //-- Save data type and data.
    node_p->data.type = MAP_TYPE_CODE;
    node_p->data.mix.scalar.MAP_TYPE_ID = arg_data;

    //-- Return success.
    return 0;
}

//-----------------------------------------------------------------------------
// function	map_xx_fetch
//
// purpose	Fetch the data from the current node in the mapping and return
//		directly as a value.  There is no means to indicate an error.
//
// arguments	1 (MAP) mapping handle
//
// return	(map_type) data value returned
//		(map_type) 0 cast to the data type if error
//-----------------------------------------------------------------------------
map_type
map_xx_fetch (
    MAP			arg_map
)
{
    struct map_node *	node_p	;

    //-- Make sure we have a mapping and a current node.
    if ( ! arg_map ) return (map_type) 0;
    if ( ! ( node_p = avl_find( & (arg_map->avl_tree), NULL, 0 ) ) ) return (map_type) 0;

    //-- Make sure data type is correct.
    if ( node_p->data.type != MAP_TYPE_CODE ) return (map_type) 0;

    //-- Return data value.
    return node_p->data.mix.scalar.MAP_TYPE_ID;
}
